# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Textual UI for text-based bots.
"""

from typing import Dict, Any
from textual import on
from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.containers import Horizontal, Vertical
from textual.widgets import Markdown, Input
from intentional_core import BotStructure


class ChatHistory(Markdown):
    """
    A markdown widget that displays the chat history.
    """


class MessageBox(Input):
    """
    An input widget that allows the user to type a message.
    """

    placeholder = "Message..."


class SystemPrompt(Markdown):
    """
    A markdown widget that displays the system prompt.
    """


class TextChatInterface(App):
    """
    The main interface class for the text-based bot UI.
    """

    CSS_PATH = "example.tcss"

    def __init__(self, bot: BotStructure):
        super().__init__()
        self.bot = bot
        self.bot.add_event_handler("on_text_message_from_llm", self.handle_text_messages)
        self.bot.add_event_handler("on_llm_starts_generating_response", self.handle_start_text_response)
        self.bot.add_event_handler("on_llm_stops_generating_response", self.handle_finish_text_response)
        self.bot.add_event_handler("on_system_prompt_updated", self.handle_system_prompt_updated)

        self.conversation = ""
        self.generating_response = False

    def compose(self) -> ComposeResult:
        """
        Layout for the text-based bot UI.
        """
        yield Horizontal(
            Vertical(
                Markdown("# Chat History"),
                ScrollableContainer(ChatHistory()),
                MessageBox(placeholder="Message..."),
                classes="column bordered chat",
            ),
            Vertical(
                Markdown("# System Prompt"),
                ScrollableContainer(SystemPrompt()),
                classes="column bordered",
            ),
        )

    def on_mount(self) -> None:
        """
        Operations to perform when the UI is mounted.
        """
        self.query_one(SystemPrompt).update(self.bot.llm.system_prompt)
        self.query_one(MessageBox).focus()

    @on(MessageBox.Submitted)
    async def send_message(self, event: MessageBox.Changed) -> None:
        """
        Sends a message to the bot when the user presses enter.

        Args:
            event: The event containing the message to send.
        """
        self.conversation += "\n\n**User**: " + event.value
        self.query_one(MessageBox).clear()
        self.query_one(ChatHistory).update(self.conversation)
        await self.bot.send({"text_message": {"role": "user", "content": event.value}})

    async def handle_start_text_response(self, _) -> None:
        """
        Prints to the console when the bot starts generating a text response.
        """
        if not self.generating_response:  # To avoid the duplication due to function calls.
            self.generating_response = True
            self.conversation += "\n\n**Assistant:** "
            self.query_one(ChatHistory).update(self.conversation)

    async def handle_finish_text_response(self, _) -> None:
        """
        Prints to the console when the bot stops generating a text response.
        """
        self.generating_response = False

    async def handle_text_messages(self, event: Dict[str, Any]) -> None:
        """
        Prints to the console any text message from the bot. It is usually a chunk as the output is being streamed out.

        Args:
            event: The event dictionary containing the message chunk.
        """
        if event["delta"]:
            self.conversation += event["delta"]
            self.query_one(ChatHistory).update(self.conversation)

    async def handle_system_prompt_updated(self, event: Dict[str, Any]) -> None:
        """
        Prints to the console any text message from the bot.

        Args:
            event: The event dictionary containing the message.
        """
        self.query_one(SystemPrompt).update(event["system_prompt"])  # self.bot.llm.system_prompt)



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Textual UI for audio stream bots.
"""

from typing import Dict, Any
import base64
import structlog
from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.containers import Horizontal, Vertical
from textual.widgets import Markdown
from intentional_core import BotStructure
from intentional_terminal.handlers import AudioHandler


log = structlog.get_logger(logger_name=__name__)


class ChatHistory(Markdown):
    """
    A markdown widget that displays the chat history.
    """


class UserStatus(Markdown):
    """
    A markdown widget that displays the user status (speaking/silent).
    """


class SystemPrompt(Markdown):
    """
    A markdown widget that displays the system prompt.
    """


class AudioStreamInterface(App):
    """
    The main interface class for the audio stream bot UI.
    """

    CSS_PATH = "example.tcss"

    def __init__(self, bot: BotStructure, audio_output_handler: "AudioHandler"):
        super().__init__()
        self.bot = bot
        self.audio_handler = audio_output_handler
        self.bot.add_event_handler("on_audio_message_from_llm", self.handle_audio_messages)
        self.bot.add_event_handler("on_llm_speech_transcribed", self.handle_transcript)
        self.bot.add_event_handler("on_user_speech_transcribed", self.handle_transcript)
        self.bot.add_event_handler("on_user_speech_started", self.handle_start_user_response)
        self.bot.add_event_handler("on_user_speech_ended", self.handle_finish_user_response)
        self.bot.add_event_handler("on_system_prompt_updated", self.handle_system_prompt_updated)
        self.bot.add_event_handler("on_conversation_ended", self.handle_conversation_end)

        self.conversation = ""

    def compose(self) -> ComposeResult:
        """
        Layout of the UI.
        """
        yield Horizontal(
            Vertical(
                Markdown("# Chat History"),
                ScrollableContainer(ChatHistory()),
                UserStatus("# User is silent..."),
                classes="column bordered chat",
            ),
            Vertical(
                Markdown("# System Prompt"),
                SystemPrompt(),
                classes="bordered column",
            ),
        )

    def on_mount(self) -> None:
        """
        Operations to be performed at mount time.
        """
        self.query_one(SystemPrompt).update(self.bot.llm.system_prompt)

    async def handle_transcript(self, event: Dict[str, Any]) -> None:
        """
        Prints the transcripts in the chat history.
        """
        if event["type"] == "on_user_speech_transcribed":
            self.conversation += f"\n**User:** {event['transcript']}\n"
        elif event["type"] == "on_llm_speech_transcribed":
            self.conversation += f"\n**Assistant:** {event['transcript']}\n"
        else:
            log.debug("Unknown event with transcript received.", event_name=event["type"])
            self.conversation += f"\n**{event['type']}:** {event['transcript']}\n"
        self.query_one(ChatHistory).update(self.conversation)

    async def handle_system_prompt_updated(self, event: Dict[str, Any]) -> None:
        """
        Prints to the console any text message from the bot.

        Args:
            event: The event dictionary containing the message.
        """
        self.query_one(SystemPrompt).update(event["system_prompt"])

    async def handle_start_user_response(self, _) -> None:
        """
        Updates the user status when they start speaking.
        """
        self.query_one(UserStatus).update("# User is speaking...")

    async def handle_finish_user_response(self, _) -> None:
        """
        Updates the user status when they stop speaking.
        """
        self.query_one(UserStatus).update("# User is silent...")

    async def handle_audio_messages(self, event: Dict[str, Any]) -> None:
        """
        Plays audio responses from the bot and updates the bot status line.

        Args:
            event: The event dictionary containing the audio message.
        """
        # self.query_one(BotStatus).update("# Bot is speaking...")
        if event["delta"]:
            self.audio_handler.play_audio(base64.b64decode(event["delta"]))

    async def handle_conversation_end(self, _) -> None:
        """
        At the end of the conversation, closes the UI.
        """
        self.exit(0)
        self.audio_handler.stop_streaming()
        self.audio_handler.cleanup()



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Local bot interface for Intentional.
"""

from typing import Any, Dict, List, Callable

import asyncio
import threading

import structlog
from intentional_core import (
    BotInterface,
    BotStructure,
    load_bot_structure_from_dict,
    IntentRouter,
)
from intentional_terminal.handlers import AudioHandler

from intentional_textual_ui.audio_stream_ui import AudioStreamInterface
from intentional_textual_ui.text_chat_ui import TextChatInterface


log = structlog.get_logger(logger_name=__name__)


class TextualUIBotInterface(BotInterface):
    """
    Bot that uses a Textual UI command line interface to interact with the user.
    """

    name = "textual_ui"

    def __init__(self, config: Dict[str, Any], intent_router: IntentRouter):
        # Init the structure
        bot_structure_config = config.pop("bot", None)
        if not bot_structure_config:
            raise ValueError(
                f"{self.__class__.__name__} requires a 'bot' configuration key to know how to structure the bot."
            )
        log.debug("Creating bot structure", bot_structure_config=bot_structure_config)
        self.bot: BotStructure = load_bot_structure_from_dict(intent_router=intent_router, config=bot_structure_config)

        # Check the modality
        self.modality = config.pop("modality")
        log.debug("Setting modality for bot structure", modality=self.modality)

        # Handlers
        self.audio_handler = None
        self.input_handler = None
        self.app = None

    async def run(self) -> None:
        """
        Chooses the specific loop to use for this combination of bot and modality and kicks it off.
        """
        if self.modality == "audio_stream":
            await self._run_audio_stream(self.bot)

        elif self.modality == "text_messages":
            await self._run_text_messages(self.bot)

        else:
            raise ValueError(
                f"Modality '{self.modality}' is not yet supported."
                "These are the supported modalities: 'audio_stream', 'text_messages'."
            )

    async def _run_text_messages(self, bot: BotStructure) -> None:
        """
        Runs the CLI interface for the text turns modality.
        """
        log.debug("Running in text turns mode.")
        self.app = TextChatInterface(bot=bot)
        await bot.connect()
        await self._launch_ui()

    async def _run_audio_stream(self, bot: BotStructure) -> None:
        """
        Runs the CLI interface for the continuous audio streaming modality.
        """
        log.debug("Running in continuous audio streaming mode.")

        try:
            self.audio_handler = AudioHandler()
            self.app = AudioStreamInterface(bot=bot, audio_output_handler=self.audio_handler)
            await bot.connect()
            await self._launch_ui(gather=[self.bot.run()])
            await self.audio_handler.start_streaming(bot.send)

        except Exception as e:  # pylint: disable=broad-except
            raise e
        finally:
            self.audio_handler.stop_streaming()
            self.audio_handler.cleanup()
            await bot.disconnect()
            print("Chat is finished. Bye!")

    async def _launch_ui(self, gather: List[Callable] = None):
        """
        Launches the Textual UI interface. If there's any async task that should be run in parallel, it can be passed
        as a list of callables to the `gather` parameter.

        Args:
            gather: A list of callables that should be run in parallel with the UI.
        """
        self.app._loop = asyncio.get_running_loop()  # pylint: disable=protected-access
        self.app._thread_id = threading.get_ident()  # pylint: disable=protected-access
        with self.app._context():  # pylint: disable=protected-access
            try:
                if not gather:
                    await self.app.run_async(
                        headless=False,
                        inline=False,
                        inline_no_clear=False,
                        mouse=True,
                        size=None,
                        auto_pilot=None,
                    )
                else:
                    asyncio.gather(
                        self.app.run_async(
                            headless=False,
                            inline=False,
                            inline_no_clear=False,
                            mouse=True,
                            size=None,
                            auto_pilot=None,
                        ),
                        *gather,
                    )
            finally:
                self.app._loop = None  # pylint: disable=protected-access
                self.app._thread_id = 0  # pylint: disable=protected-access



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Init file for the intentional_textual_ui package.
"""
from intentional_textual_ui.bot_interface import TextualUIBotInterface

__all__ = ["TextualUIBotInterface"]
